#!/usr/bin/env python3

# Python script for BW admin tools

# External module required: 
# pip3 install requests
# 
# Requirements Before Running The Script
# - Fill in the config file. See config-example.cfg


import json
import requests
from requests.exceptions import Timeout
import time
import sys
import getopt
import configparser
import os
import re
import subprocess
import getpass
import shutil

#TODO: Automatic run of CLI and bw serve?
bw_vault_uri = ""
bw_identity_endpoint = ""
bw_api_endpoint = ""
bw_org_client_id = ""
bw_org_client_secret = ""
bw_vault_mgmt_uri = ""
bw_org_id=""
bw_acc_client_id = ""
bw_acc_client_secret = ""
bw_acc_password = ""

dest_bw_vault_uri = ""
dest_bw_org_client_id = ""
dest_bw_org_client_secret = ""
dest_bw_acc_client_id = ""
dest_bw_acc_client_secret = ""
dest_bw_org_id=""
dest_bw_identity_endpoint = ""
dest_bw_api_endpoint = ""
dest_bw_acc_password = ""

lp_cid = ""
lp_api_secret = ""
lp_api_uri = ""

access_token = ""
dest_access_token = ""
bw_path = ""

delay_after_api_call_secs = 1
debug = False
verbose = False
add_group_if_not_exists = True

script_location = os.path.dirname(os.path.abspath(__file__))


def delete_all_collections(f_bw_identity_endpoint, f_bw_api_endpoint, f_bw_org_client_id, f_bw_org_client_secret):
    global access_token
    if access_token == "":
        access_token = login_to_bw_public_api(f_bw_identity_endpoint,f_bw_org_client_id,f_bw_org_client_secret)

    http_headers = {'Authorization': 'Bearer '+access_token}
    response = requests.get(bw_api_endpoint+"public/collections", headers=http_headers)
    

    if (response.status_code == 200):
        col_list = response.json()
        if len(col_list["data"]) > 0:
            for collection in col_list["data"]:
                response = requests.delete(f_bw_api_endpoint+"public/collections/"+collection["id"], headers=http_headers)
                if (response.status_code == 200):
                    if verbose: print("Collection ID: "+collection["id"]+" deleted")
                else:
                    if verbose: print("Collection ID: "+collection["id"]+" failed with status: ",response.status_code)
                time.sleep(delay_after_api_call_secs)
        else:
            if verbose: print("Empty Collection")
    else:
        if verbose: print("Getting Collections failed with status code: "+response.status_code )

def delete_all_groups(f_bw_identity_endpoint, f_bw_api_endpoint, f_bw_org_client_id, f_bw_org_client_secret):
    global access_token
    if access_token == "":
        access_token = login_to_bw_public_api(f_bw_identity_endpoint,f_bw_org_client_id,f_bw_org_client_secret)

    http_headers = {'Authorization': 'Bearer '+access_token}
    response = requests.get(f_bw_api_endpoint+"public/groups", headers=http_headers)
    

    if (response.status_code == 200):
        group_list = response.json()
        if len(group_list["data"]) > 0:
            for group in group_list["data"]:
                response = requests.delete(f_bw_api_endpoint+"public/groups/"+group["id"], headers=http_headers)
                if (response.status_code == 200):
                    if verbose: print("Group ID: "+group["id"]+" deleted")
                else:
                    if verbose: print("Group ID: "+group["id"]+" failed with status: ",response.status_code)
                time.sleep(delay_after_api_call_secs)
        else:
            if verbose: print("Empty Groups")
    else:
        if verbose: print("Getting Groups failed with status code: "+response.status_code )

def load_collections_for_lastpass(bw_vault_mgmt_uri,bw_org_id):
    # Function to load all collections from vault management API to a dictionary
    http_headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
    response = requests.get(bw_vault_mgmt_uri+"list/object/org-collections?organizationId="+bw_org_id, headers=http_headers)

    collection_dict = {}

    if (response.status_code == 200):
        collection_list = response.json()
        if len(collection_list["data"]) == 0 or len(collection_list["data"]["data"]) == 0:
            print("Collection list is empty")
        else:
            for collection in collection_list["data"]["data"]:
                collection_dict[collection["name"]] = collection["id"]
    else:
        print("Getting Collection List Failed. Status Code: ",response.status_code)

    return collection_dict

def load_groups_api(f_bw_api_endpoint,f_access_token):
    # Function to load all groups from public API to a dictionary
    http_headers = {'Authorization': 'Bearer '+f_access_token}
    response = requests.get(f_bw_api_endpoint+"public/groups", headers=http_headers)
    
    group_data = {}
    if (response.status_code == 200):
        group_json = response.json()
        if "data" in group_json:
            group_data = group_json["data"]
        else:
            print(f"JSON Response does not contain data. Response: {response}")
            sys.exit(2)
    else:
        print(f"Failed loading groups list. Response: {response.status_code}")
        sys.exit(2)

    if (debug):
        print("Load groups debug")
        print("API")
        print(f_bw_api_endpoint)
        print("Response")
        print(response)
        print("")

    return group_data

def load_group_details_api(f_bw_api_endpoint,f_access_token, group_id):
    # Function to load all groups from public API to a dictionary
    http_headers = {'Authorization': 'Bearer '+f_access_token}
    response = requests.get(f_bw_api_endpoint+"public/groups/"+group_id, headers=http_headers)
    
    group_data = {}
    if (response.status_code == 200):
        group_json = response.json()
        if len(group_json) > 0:
            group_data = group_json
        else:
            print(f"JSON Response does not contain data. load_group_details_api. Response: {response}. Group ID:{group_id}")
    else:
        print(f"Failed loading group details. Response: {response.status_code} Group ID:{group_id}")
        sys.exit(2)

    return group_data

def load_collections_api(f_bw_api_endpoint,f_access_token):
    # Function to load all collections from public API to a dictionary
    http_headers = {'Authorization': 'Bearer '+f_access_token}
    response = requests.get(f_bw_api_endpoint+"public/collections", headers=http_headers)
    
    f_coll_dict = {}
    if (response.status_code == 200):
        coll_list = response.json()
        if len(coll_list["data"]) > 0:
            f_coll_dict =  coll_list["data"]
    return f_coll_dict

def load_collections_cli(f_bw_org_id, f_bw_cli_session):

    output = subprocess.check_output([bw_path, 'list', 'org-collections', '--organizationid', f_bw_org_id, '--session', f_bw_cli_session])

    output_str = output.decode('utf-8')
    data = json.loads(output_str)

    return data

def get_members_list(f_bw_api_endpoint,f_access_token):
    # Function to load all groups from public API to a dictionary
    http_headers = {'Authorization': 'Bearer '+f_access_token}
    response = requests.get(f_bw_api_endpoint+"public/members", headers=http_headers)
    
    f_member_list = []
    if (response.status_code == 200):
        resp_dict = response.json()
        f_member_list = resp_dict["data"]
    else:
        print(f"Failed getting member list. Response Code: {response.status_code}")

    if (debug):
        print("Load members debug")
        print("API")
        print(f_bw_api_endpoint)
        print("Response")
        print(response)
        print("BW Member List")
        print(f_member_list)
        print("")

    return f_member_list

def create_a_group(f_bw_api_endpoint, f_access_token, f_group_name, f_external_id):
    # Function to create a group via BW Public API

    json_body = { "name": f_group_name, "accessAll": False, "externalId": f_external_id }
    http_headers = {'Authorization': 'Bearer ' + f_access_token}
    response = requests.post(f_bw_api_endpoint+"public/groups", json=json_body, headers=http_headers)
    
    if (response.status_code == 200):
        json_resp = response.json()
    else:
        print("Error creating a group. Response Code: " + response.status_code)
        json_resp = {}
    return json_resp

def update_collection(bw_vault_mgmt_uri,bw_org_id,collection_id,json_body):
    # Function to update the permissions of a collection using vault management API

    http_headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
    response = requests.put(bw_vault_mgmt_uri+"object/org-collection/"+collection_id+"?organizationId="+bw_org_id, json=json_body, headers=http_headers)
    if (response.status_code == 200):
        if (debug): print("Editing Collection Successful. ID: ",collection_id)
    else:
        print("Editing Collection Failed: ",collection_id, " | Status Code: ",response.status_code)
        if (debug):
            print("JSON Body")
            print(json.dumps(json_body))
            print("")

    return response.status_code
    
def login_to_bw_public_api(f_bw_identity_endpoint, f_bw_org_client_id, f_bw_org_client_secret):
    #Function to log in to BW Public API and get access token

    f_access_token = ""
    body = {"client_id":f_bw_org_client_id,"client_secret":f_bw_org_client_secret,"grant_type":"client_credentials","scope":"api.organization"}
    http_headers = {'Content-type': 'application/x-www-form-urlencoded'}

    try:
        response = requests.post(f_bw_identity_endpoint+"connect/token", data=body, headers=http_headers, timeout = 10)
    except Timeout:
        print("The request timed out during authentication")
        print("Identity endpoint: ", f_bw_identity_endpoint)
        exit(1)
    except Exception as e:
        print("An error occurred during auth to public API: ", e)
        print("Identity endpoint: ", f_bw_identity_endpoint)
        exit(1)

    if (response.status_code == 200):
        response_json = response.json()
        f_access_token = response_json["access_token"]
    else:
        f_access_token = ""
        if (debug): print(response.content )
        print("Auth to BW Public API failed, Status Code: ",response.status_code, " Endpoint: ", f_bw_identity_endpoint)

    if f_access_token == "":
        sys.exit(2)
    return f_access_token

def get_lastpass_shared_folders(cid, provhash, lp_api_uri):
    # Function for getting list of shared folders from Lastpass

    body = {"cid":cid,"provhash":provhash,"cmd":"getsfdata","data":"all"}
    http_headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}

    response = requests.post(lp_api_uri, json=body, headers=http_headers)
    if (response.status_code == 200):
        shared_folders = response.json()
        if (debug):
            print("LP API Success")
            print(shared_folders)
            print("")
    else:
        print("Failed getting Shared Folder data from Lastpass",response.status_code)
        shared_folders = []
    return shared_folders

def update_user_collection( f_bw_api_endpoint, f_access_token, f_user_id, f_data_dict ):

    #http_headers = {'Authorization': 'Bearer ' + f_access_token}
    #json_body = { "type": user_type, 'accessAll': user_accessAll, 'externalId': user_externalId, 'resetPasswordEnrolled': user_resetPasswordEnrolled, "collections": user_collections }

    http_headers = {'Authorization': 'Bearer ' + f_access_token}
    response = requests.put(f_bw_api_endpoint + "public/members/" + f_user_id, json=f_data_dict, headers=http_headers)
    
    if (response.status_code == 200):
        if(verbose): print("Updating collection for ",f_user_id," successful")
    else:
        if (verbose): print("Error updating user ",f_user_id," Status Code: ",response.status_code)
    time.sleep(delay_after_api_call_secs)

def update_user_collection_old_lastpass( user_id, user_name, user_type, user_accessAll, user_externalId, user_resetPasswordEnrolled, user_collections ):
    global access_token
    if access_token == "":
        access_token = login_to_bw_public_api(bw_identity_endpoint,bw_org_client_id,bw_org_client_secret)

    http_headers = {'Authorization': 'Bearer '+access_token}


    json_body = { "type": user_type, 'accessAll': user_accessAll, 'externalId': user_externalId, 'resetPasswordEnrolled': user_resetPasswordEnrolled, "collections": user_collections }
    http_headers = {'Authorization': 'Bearer ' + access_token}
    response = requests.put(bw_api_endpoint+"public/members/"+user_id, json=json_body, headers=http_headers)
    
    if (response.status_code == 200):
        json_resp = response.json()
        if(verbose): print("Updating collection for ",user_name," successful")
    else:
        if (verbose): print("Error updating user ",user_name," Status Code: ",response.status_code)
    time.sleep(delay_after_api_call_secs)

def update_all_individual_accounts(account_list):
    global access_token
    if access_token == "":
        access_token = login_to_bw_public_api(bw_identity_endpoint,bw_org_client_id,bw_org_client_secret)

    http_headers = {'Authorization': 'Bearer '+access_token}
    response = requests.get(bw_api_endpoint+"public/members", headers=http_headers)
    
    if (response.status_code == 200):
        json_resp = response.json()
        if (debug):
            print("All User Data:\n",json_resp)
        if len(json_resp['data']) > 0:
            for user in json_resp['data']:
                if user["email"] in account_list:
                    #this is not working because collections is always None from api/public/members
                    if user["collections"] is None:
                        user["collections"] = account_list[user["email"]]
                    else:
                        user["collections"].append(account_list[user["email"]])

                    update_user_collection_old_lastpass(user["id"], user["name"], user["type"], user["accessAll"], user["externalId"], user["resetPasswordEnrolled"], user["collections"])
        else:
            if (debug or verbose): print("User list is empty")
    else:
        if (debug): print("Error getting all users, status: ", response.status_code)
        json_resp = {}
    return True

def check_vault_mgmt_server(bw_vault_mgmt_uri):

    http_headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}

    try:
        response = requests.get(bw_vault_mgmt_uri+"sync", headers=http_headers)
    except requests.exceptions.Timeout:
        # Maybe set up for a retry, or continue in a retry loop
        print("Connection to Vault Management API failed. Please make sure the Vault Management API is running")
        sys.exit(2)
    except requests.exceptions.ConnectionError as errc:
        print("Connection to Vault Management API",bw_vault_mgmt_uri,"failed. Please make sure the Vault Management API is running")
        sys.exit(2)
    except requests.exceptions.RequestException as e:
        # catastrophic error. bail.
        raise SystemExit(e)

def migrate_shared_folders_permissions():
    global access_token
    if access_token == "":
        access_token = login_to_bw_public_api(bw_identity_endpoint,bw_org_client_id,bw_org_client_secret)

    check_vault_mgmt_server(bw_vault_mgmt_uri)

    collection_dict = load_collections_for_lastpass(bw_vault_mgmt_uri,bw_org_id)

    if len(collection_dict) == 0:
        print("Collection Empty or Loading BW Collections failed, exiting program")
        sys.exit(2)


    #group_dict, group_dict_id = load_groups(bw_api_endpoint,access_token)
    #del group_dict_id

    group_list = load_groups_api(bw_api_endpoint,access_token)
    group_dict_name = {}
    for group in group_list:
        group_dict_name[group["name"]] = {"id" :group["id"], "name":group["name"], "externalId":group["externalId"]}

 
    if not add_group_if_not_exists and len(group_dict_name) == 0:
        print("Groups Empty or Loading BW Groups failed, exiting program")
        sys.exit(2)


    #Getting Shared Folders data from Lastpass
    #Enable this when trying real data
    
    shared_folders = get_lastpass_shared_folders(lp_cid, lp_api_secret, lp_api_uri)

    # with open("sample.json", "w") as outfile:
    #     outfile.write( json.dumps(shared_folders) )

    # read from file so I don't have to flood LP API during development and risk getting blocked
    # f = open('sample.json')
    # shared_folders = json.load(f)

    if (debug):
        print("Lastpass JSON\n",shared_folders)

    if len(shared_folders) == 0:
        print("No shared folders data. Exiting program")
        sys.exit(2)
    
    account_list = {}

    #loop every shared folders from lastpass API
    for folder in shared_folders:
        if (debug): print("Processing folder:",folder)
        if (shared_folders[folder]["sharedfoldername"] in collection_dict.keys()) or ("Shared-"+shared_folders[folder]["sharedfoldername"] in collection_dict.keys()):
            if shared_folders[folder]["sharedfoldername"] in collection_dict.keys():
                shared_folder_name = shared_folders[folder]["sharedfoldername"]
            else:
                shared_folder_name = "Shared-" + shared_folders[folder]["sharedfoldername"]

            if not shared_folders[folder]["deleted"] or shared_folders[folder]["deleted"] == "False":
                json_body = {"organizationId": bw_org_id, "name": shared_folder_name,"externalid": "null"}
                if (debug): print(json_body)
                json_group = []
                group_added = {}
                #loop all users in the shared folder
                for user in shared_folders[folder]["users"]:
                    if ("group_name" in user):
                        #Create the group if group doesnt exists in Bitwarden
                        if not user["group_name"] in group_dict_name:
                            if add_group_if_not_exists:
                                json_resp = create_a_group(bw_api_endpoint, access_token, user["group_name"], '' )
                                group_dict_name[user["group_name"]] = {"id" :json_resp["id"], "name":json_resp["name"], "externalId":json_resp["externalId"]}
                                if (debug):
                                    print("Group Added: ",user["group_name"])

                        if not user["group_name"] in group_added and user["group_name"] in group_dict_name:
                            group_added[user["group_name"]] = True
                            if (debug): print("username: ",user["username"],"group: ",type(user["group_name"]))
                            if user["readonly"] == "1":
                                readOnly = True
                            else:
                                readOnly = False

                            if user["give"] == "1":
                                hidePasswords = False
                            else:
                                hidePasswords = True

                            json_group.append({  "id": group_dict_name[user["group_name"]]["id"], "readOnly": readOnly, "hidePasswords": hidePasswords })
                    else:
                        # If group name not found, this is individual user permission
                        if (debug): print("add individual ",shared_folder_name,"| username: ",user["username"])
                        if user["username"] not in account_list:
                            account_list[user["username"]] = []
                        if user["readonly"] == "1":
                            readOnly = True
                        else:
                            readOnly = False

                        if user["give"] == "1":
                            hidePasswords = False
                        else:
                            hidePasswords = True
                        account_list[user["username"]].append({ "id": collection_dict[shared_folder_name], "readOnly": readOnly  })

                json_body["groups"] = json_group
                update_collection(bw_vault_mgmt_uri,bw_org_id,collection_dict[shared_folder_name],json_body)
                time.sleep(delay_after_api_call_secs)
                if (debug):
                    print("Group List For Collection ID: ",collection_dict[shared_folder_name])
                    print(json_group)
    #Update all individual accounts affected
    update_all_individual_accounts(account_list)

def find_bw():
    # Check if 'bw' file exists in the same directory as the script
    script_dir = os.path.dirname(os.path.realpath(__file__))
    bw_path = os.path.join(script_dir, 'bw')

    if os.path.isfile(bw_path):
        # Check if 'bw' is executable
        if os.access(bw_path, os.X_OK):
            return bw_path
        else:
            print(f"'bw' file at {bw_path} is not executable")

    # If 'bw' file not found in script directory, or it's not executable
    # check if 'bw' is a system-wide command
    system_bw = shutil.which('bw')
    if system_bw:
        return system_bw

    return ""

def initial_environment_check():
    global bw_path
    # meant for checking things
    # maybe download the CLI?
    is_writable = os.access(script_location, os.W_OK)
    if not is_writable:
        print("The script's location is not writable.")
        sys.exit(1)

    bw_path = find_bw()
    if bw_path == "":
        print("Bitwarden CLI (bw) is not found in the system")
        sys.exit(2)

def get_account_passwords():
    global bw_acc_password, dest_bw_acc_password

    bw_acc_password = getpass.getpass("Please input your Bitwarden password used on the source server:")
    dest_bw_acc_password = getpass.getpass("Please input your Bitwarden password for the destination server:")

    if (bw_acc_password == "") or (dest_bw_acc_password == ""):
        print("Both passwords are required to migrate your encrypted data")
        sys.exit(1)

def delete_file(filepath):
    if os.path.isfile(filepath):
        try:
            os.remove(filepath)
        except OSError as e:
            print(f"Error: {e.filename} - {e.strerror}.")

def delete_all_export_files():
    delete_file(os.path.join(script_location, "export.json"))

    dir_path =  os.path.join(script_location , "attachments")
    if os.path.isdir(dir_path):
        try:
            shutil.rmtree(dir_path)
        except OSError as e:
            print(f"Error: {e.filename} - {e.strerror}.")
    else:
        print(f"Error: {dir_path} does not exist.")

def login_on_cli(f_bw_vault_uri, f_bw_acc_client_id, f_bw_acc_client_secret, f_bw_acc_password):
    f_cli_session = ""

    os.environ["BW_CLIENTID"] = f_bw_acc_client_id
    os.environ["BW_CLIENTSECRET"] = f_bw_acc_client_secret
    os.environ["BW_PASSWORD"] = f_bw_acc_password

    output = subprocess.check_output([bw_path, 'status'])
    output_str = output.decode('utf-8')
    data = json.loads(output_str)

    if not data["status"] == "unauthenticated":
        subprocess.run([bw_path, 'logout', '--raw'])
    if f_bw_vault_uri == "https://vault.bitwarden.com/":
        subprocess.run([bw_path, 'config', 'server', 'null', '--raw'])
    else:
        subprocess.run([bw_path, 'config', 'server', f_bw_vault_uri, '--raw'])

    os.environ["BW_CLIENTID"] = f_bw_acc_client_id
    os.environ["BW_CLIENTSECRET"] = f_bw_acc_client_secret
    try:        
        subprocess.run([bw_path, 'login', '--apikey', '--raw'])
    except subprocess.CalledProcessError as e:
        print(f"There is an issue logging in to your origin server. Exit code: {e.returncode}\nError: {e.stderr}")
        exit(1)


    try:
        output = subprocess.check_output([bw_path, 'unlock', '--passwordenv', 'BW_PASSWORD', '--raw'])
        output_str = output.decode('utf-8')
    except subprocess.CalledProcessError as e:
        print(f"There is an issue unlocking your vault with client id {f_bw_acc_client_id}. Make sure you entered the correct password. Exit code: {e.returncode}\nError: {e.stderr}")
        exit(1)

    f_cli_session = output_str
    os.environ["BW_PASSWORD"] = ""
    os.environ["BW_CLIENTSECRET"] = ""

    return f_cli_session

def export_data_from_origin(f_bw_cli_session):

    # export all attachments

    attach_dir_path = os.path.join(script_location ,"attachments")

    try:
        os.mkdir(attach_dir_path)
    except OSError as e:
        print(f"Error: {e.strerror}.")

    command = [bw_path, "list", "items", "--organizationid", bw_org_id, "--session", f_bw_cli_session]
    output = subprocess.check_output(command)

    output_str = output.decode('utf-8')
    data = json.loads(output_str)
    new_created_coll = []

    for item in data:
        if "attachments" in item:

            #create a collection for that item
            new_col_dict = {"organizationId" :bw_org_id, "externalid": None, "name": item['id'], "groups": [] }

            new_coll_data = create_collection_cli(f_bw_cli_session, item['id'], new_col_dict, bw_org_id)
            item["collectionIds"].append(new_coll_data["id"])
            new_created_coll.append(new_coll_data["id"])

            input_data = json.dumps(item["collectionIds"]).encode()

            cmd3 = [bw_path, 'encode']
            cmd4 = [bw_path,  'edit', 'item-collections', item['id'], '--organizationid', bw_org_id, '--session', f_bw_cli_session,]

            p3 = subprocess.Popen(cmd3, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
            output3 = p3.communicate(input=input_data)[0]

            p4 = subprocess.Popen(cmd4, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
            output, _ = p4.communicate(input=output3)


            attachment_dir = os.path.join(attach_dir_path, item["id"])
            
            try:
                os.mkdir(attachment_dir)
            except OSError as e:
                print(f"Error: {e.strerror}.")

            for attachment in item["attachments"]:
                #print("output of attachment:", os.path.join(attachment_dir, attachment['fileName']))
                subprocess.run([
                    bw_path, 
                    "get", 
                    "attachment", 
                    attachment['fileName'],
                    "--itemid", 
                    str(item['id']), 
                    "--output", 
                    os.path.join(attachment_dir, attachment['fileName']),
                    "--session", 
                    f_bw_cli_session,
                    "--raw"
                ], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)

    subprocess.run([bw_path, 'sync', '--session', f_bw_cli_session, '--raw'])

    subprocess.run([
        bw_path, 
        "export",
        "--organizationid",
        str(bw_org_id),
        "--output",
        os.path.join(script_location,'export.json'),
        "--format",
        "json",
        "--session", 
        f_bw_cli_session,
        "--raw"
    ], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
    
def import_attachments_to_destination(f_dest_bw_cli_session, f_new_coll_name):
    global dest_bw_org_id

    # specify the directory you want to check
    directory_path = os.path.join(script_location , "attachments")

    # check if the directory exists
    if os.path.isdir(directory_path):        
        # list all the directories inside it
        all_subdirectories = [name for name in os.listdir(directory_path) 
                            if os.path.isdir(os.path.join(directory_path, name))]
        
        # list all files in the subdirectories
        for subdir in all_subdirectories:

            #get the collection id
            new_coll_id = f_new_coll_name[subdir]
            output = subprocess.check_output([bw_path, 'list', 'items', '--collectionid', new_coll_id, '--session', f_dest_bw_cli_session])

            output_str = output.decode('utf-8')
            new_item_arr = json.loads(output_str)
            if len(new_item_arr) > 0:
                new_item_dict = new_item_arr.pop()
                subdir_path = os.path.join(directory_path, subdir)
                for dirpath, dirnames, filenames in os.walk(subdir_path):
                    for filename in filenames:
                        full_path_filename = os.path.join(subdir_path, filename)
                        subprocess.run([
                            bw_path, 
                            "create", 
                            "attachment",
                            "--file",
                            full_path_filename,
                            "--itemid", 
                            str(new_item_dict['id']),
                            "--session", 
                            f_dest_bw_cli_session,
                            "--raw"
                        ], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
                        
            else:
                print(f"item with collection id {new_coll_id} is not found. unable to import attachment")
            
            #delete collection
            subprocess.run([
                bw_path, 
                "delete", 
                "org-collection",
                "--organizationid",
                str(dest_bw_org_id),
                new_coll_id,
                "--session", 
                f_dest_bw_cli_session
            ])
            
def import_data_to_destination(f_dest_bw_cli_session, f_new_coll_name):
    global dest_bw_org_id, script_location

    # import the JSON file
    subprocess.run([
        bw_path, 
        "import", 
        "bitwardenjson",
        "--organizationid",
        str(dest_bw_org_id),
        os.path.join(script_location, "export.json"),
        "--session", 
        f_dest_bw_cli_session,
        "--raw"
    ])
    subprocess.run([
        bw_path, 
        "sync",
        "--session", 
        f_dest_bw_cli_session,
        "--raw"
    ])

    import_attachments_to_destination(f_dest_bw_cli_session, f_new_coll_name)

    #clean up the files
    delete_all_export_files()    

def is_uuid(string):
    uuid_regex = r'^[0-9a-f]{8}-[0-9a-f]{4}-[1-5][0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}\b'
    match = re.match(uuid_regex, string)
    return bool(match)

def load_permissions_from_origin(f_bw_cli_session):
    global access_token

    coll_dict = load_collections_cli(bw_org_id, f_bw_cli_session)

    f_coll_list = {}

    if len(coll_dict) > 0:
        if access_token == "":
            access_token = login_to_bw_public_api(bw_identity_endpoint,bw_org_client_id,bw_org_client_secret)
        
        #group_dict_name, group_dict_id = load_groups(bw_api_endpoint,access_token)
        #del group_dict_name

        group_list = load_groups_api(bw_api_endpoint,access_token)
        group_dict_id = {}
        for group in group_list:
            group_dict_id[group["id"]] = {"id" :group["id"], "name":group["name"], "externalId":group["externalId"]}


        for bw_col in coll_dict:

            # getting collection permissions details for groups

            http_headers = {'Authorization': 'Bearer ' + access_token}
            response = requests.get(bw_api_endpoint+"public/collections/" + bw_col["id"], headers=http_headers)
            
            group_permissions = []

            if (response.status_code == 200):
                col_details = response.json()
                if len(col_details["groups"]) > 0:
                    for each_group in col_details["groups"]:
                        each_group["name"] = group_dict_id[each_group["id"]]["name"]
                        group_permissions.append(each_group)

            f_coll_list[bw_col["id"]] = { "id": bw_col["id"] , "name": bw_col["name"], "groups": group_permissions, "users": []}

            #if collection name is UUID, delete it from source
            if is_uuid(bw_col["name"]):
                #delete the collection for clean up
                subprocess.run([bw_path, 'delete', 'org-collection', bw_col['id'], '--organizationid', bw_org_id, '--session', f_bw_cli_session, '--raw'])

        
        # getting collection permissions details for users

        member_list = get_members_list(bw_api_endpoint,access_token)
        f_member_details_list = []
        if len(member_list) > 0:
            for member in member_list:
                http_headers = {'Authorization': 'Bearer ' + access_token}
                response = requests.get(bw_api_endpoint+"public/members/" + member["id"], headers=http_headers)
                if (response.status_code == 200):
                    member_details = response.json()
                    if len(member_details["collections"]) > 0:
                        for collection in member_details["collections"]:
                            collection["name"] = f_coll_list[collection["id"]]["name"]
                            user_permission = {"email" : member_details["email"], "readOnly" : collection["readOnly"]}
                            f_coll_list[collection["id"]]["users"].append(user_permission)
                            f_member_details_list.append(member_details)

    return f_coll_list, f_member_details_list, group_dict_id

def create_collection_cli(f_dest_bw_cli_session, f_coll_name, f_new_data_col, f_bw_org_id):
    # Adding a collection via CLI

    #json_data = json.dumps(f_new_data_col)
    #print("creating new collection: ",json_data)

    try:
        cmd3 = [bw_path, 'encode']
        cmd4 = [bw_path, '--session', f_dest_bw_cli_session, 'create', 'org-collection', '--organizationid', f_bw_org_id]

        p3 = subprocess.Popen(cmd3, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        output3 = p3.communicate(input=json.dumps(f_new_data_col).encode())[0]

        p4 = subprocess.Popen(cmd4, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        output, _ = p4.communicate(input=output3)

        output_str = output.decode('utf-8')
        data = json.loads(output_str)
        

    except subprocess.CalledProcessError as e:
        print(f"There is an issue when creating collection {f_coll_name}. Exit code: {e.returncode}\nError: {e.stderr}")
        exit(1)

    return data

def import_permissions_to_dest(f_coll_list, f_member_details_list, f_group_dict_id, f_dest_bw_cli_session):
    global dest_access_token, dest_bw_org_id

    if dest_access_token == "":
        dest_access_token = login_to_bw_public_api(dest_bw_identity_endpoint, dest_bw_org_client_id, dest_bw_org_client_secret)

    new_group_id = {}
    if len(f_group_dict_id) > 0:
        for each_group in f_group_dict_id.values():
            json_resp = create_a_group(dest_bw_api_endpoint, dest_access_token, each_group["name"], each_group["externalId"])
            if len(json_resp) > 0:
                new_group_id[each_group["id"]] = json_resp["id"]
            else:
                print(f"Error creating group {each_group['name']}. Exiting program..")
                exit(2)
    
    if len(f_coll_list) > 0:
        new_coll_id = {}
        new_coll_name = {}
        for each_col in f_coll_list.values():
            group_array = []
            if len(each_col['groups']) > 0:
                for each_group in each_col['groups']:
                    group_array.append( {"id": new_group_id[each_group['id']], "readOnly": each_group['readOnly'], "hidePasswords": False } )

            new_data_col = {"organizationId" :dest_bw_org_id, "externalId": each_col['id'], "name": each_col['name'], "groups": group_array }

            new_coll_details = create_collection_cli(f_dest_bw_cli_session, each_col['name'], new_data_col, dest_bw_org_id)
            new_coll_id[each_col["id"]] = new_coll_details["id"]
            new_coll_name[each_col["name"]] = new_coll_details["id"]

    if len(new_coll_id) > 0:
        rewrite_export_file(new_coll_id)

    return new_coll_name        

def rewrite_export_file(f_new_coll_id):
    file_name = 'export.json'
    data = {}

    try:
        # Open the JSON file
        with open(file_name, 'r', encoding='utf-8') as json_file:
            # Load the data from the JSON file
            data = json.load(json_file)
    except FileNotFoundError:
        print(f"Error: The file {file_name} does not exist.")
        exit(2)

    except json.JSONDecodeError:
        print(f"Error: The file {file_name} is not a valid JSON file.")
        exit(2)

    except Exception as e:
        print(f"An unexpected error occurred: {e}")
        exit(2)
    
    if "collections" in data:
        if len(data["collections"]) > 0:
            for each_collection in data["collections"]:
                each_collection["id"] = f_new_coll_id[each_collection["id"]]

    if "items" in data:
        if len(data["items"]) > 0:
            for each_item in data["items"]:
                if "collectionIds" in each_item:
                    if len(each_item["collectionIds"]) > 0:
                        tmp_col = []
                        for each_collection in each_item["collectionIds"]:
                            tmp_col.append(f_new_coll_id[each_collection])
                        each_item["collectionIds"] = tmp_col

    file_name = 'export.json'
    try:
        # Open the JSON file in write mode, which will overwrite any existing file
        with open(file_name, 'w', encoding='utf-8') as json_file:
            # Write the data to the JSON file
            json.dump(data, json_file, indent=4)

    except Exception as e:
        print(f"Error writing new export file: {e}")
        exit(2)

def migrate_bw2bw_roles():
    return 0

def migrate_data_bw_to_bw():
    global bw_vault_uri, bw_acc_client_id, bw_acc_client_secret, bw_acc_password

    initial_environment_check()
    get_account_passwords()
    bw_cli_session = login_on_cli(bw_vault_uri, bw_acc_client_id, bw_acc_client_secret, bw_acc_password)
    print("Exporting data from source server...")
    export_data_from_origin(bw_cli_session)

    collection_list, member_details_list, group_dict_id = load_permissions_from_origin(bw_cli_session)

    print("Importing data to destination server...")
    dest_bw_cli_session = login_on_cli(dest_bw_vault_uri, dest_bw_acc_client_id, dest_bw_acc_client_secret, dest_bw_acc_password)

    new_coll_name = import_permissions_to_dest(collection_list, member_details_list, group_dict_id, dest_bw_cli_session)
    import_data_to_destination(dest_bw_cli_session, new_coll_name)    

def get_members_details(f_bw_api_endpoint, f_access_token, f_member_id):

    http_headers = {'Authorization': 'Bearer ' + f_access_token}
    response = requests.get(f_bw_api_endpoint + "public/members/" + f_member_id, headers=http_headers)
    member_details = {}
    if (response.status_code == 200):
        json_data = response.json()
        if len(json_data) > 0:
            member_details = json_data
        else:
            print(f"JSON Response does not contain data. load_group_details_api. Response: {response}. Group ID:{f_member_id}")
    else:
        print(f"Failed getting member details. MemberID: {f_member_id}, Response: {response.status_code}")
        sys.exit(2)

    return member_details

def load_users_details_from_api(f_bw_api_endpoint, f_access_token):

    member_list = get_members_list(f_bw_api_endpoint, f_access_token)

    f_member_details_list = {}
    if len(member_list) > 0:
        for member in member_list:
            f_member_details_list[member['email']]  = get_members_details(f_bw_api_endpoint, f_access_token, member["id"])

    return f_member_details_list

def import_users_to_dest(f_bw_api_endpoint, f_access_token, f_member_details_list):
    dest_member_list = get_members_list(f_bw_api_endpoint, f_access_token)

    coll_list = load_collections_api(f_bw_api_endpoint, f_access_token)

    #create a dictionary based on externalId. The externalId containds the collection id of source server
    coll_dict_externalid = {}
    if len(coll_list) > 0:
        for each_col in coll_list:
            coll_dict_externalid[each_col["externalId"]] = each_col["id"]


    for each_dest_member in dest_member_list:
        if each_dest_member["email"] in f_member_details_list:
            origin_user_details = f_member_details_list[each_dest_member["email"]]
            if len(origin_user_details["collections"]) > 0:
                for each_col in origin_user_details["collections"]:
                    if each_col["id"] in coll_dict_externalid:
                        each_col["id"] = coll_dict_externalid[each_col["id"]]
            user_dict = {
                "type": origin_user_details["type"],
                "accessAll": origin_user_details["accessAll"],
                "externalId": origin_user_details["externalId"],
                "resetPasswordEnrolled": each_dest_member["resetPasswordEnrolled"],
                "collections": origin_user_details["collections"]
            }
            update_user_collection( f_bw_api_endpoint, f_access_token, each_dest_member['id'], user_dict )

def migrate_users_bw_to_bw():
    global access_token, bw_api_endpoint, dest_access_token, dest_bw_api_endpoint

    initial_environment_check()

    if access_token == "":
        access_token = login_to_bw_public_api(bw_identity_endpoint, bw_org_client_id, bw_org_client_secret)

    if dest_access_token == "":
        dest_access_token = login_to_bw_public_api(dest_bw_identity_endpoint, dest_bw_org_client_id, dest_bw_org_client_secret)

    print("Exporting data from source server...")
    member_details_list = load_users_details_from_api( bw_api_endpoint, access_token)

    print("Importing data to destination server...")
    import_users_to_dest(dest_bw_api_endpoint, dest_access_token, member_details_list)

def do_diff_users(f_access_token, f_dest_access_token):

    member_details_list = load_users_details_from_api( bw_api_endpoint, f_access_token)
    dest_member_details_list = load_users_details_from_api( dest_bw_api_endpoint, f_dest_access_token)

    if len(member_details_list) < 1:
        print(f"The member list from source is empty")
    if len(dest_member_details_list) < 1:
        print(f"The member list from destination is empty")

    #will do later. Just realized that comparing users may not be required.

    return 0

def compare_dicts(dict1, dict2):
    print(f"Number of records: Source: {len(dict1)}, Destination: {len(dict2)}")

    for key in dict1.keys() | dict2.keys():
        if key not in dict1:
            print(f'"{key}" is present in destination server, but not in source')
            if len(dict2[key]) > 0:
                print('Values:', dict2[key])
        elif key not in dict2:
            print(f'"{key}" is present in source server, but not in destination')
            if len(dict1[key]) > 0:
                print('Values:', dict1[key])
        else:
            dict1_values = set(dict1[key])
            dict2_values = set(dict2[key])
            
            common_values = dict1_values & dict2_values
            only_in_dict1 = dict1_values - dict2_values
            only_in_dict2 = dict2_values - dict1_values
            
            if only_in_dict1 or only_in_dict2:
                print(f'Differences in key "{key}":')
                
                if common_values:
                    print('Common values:', list(common_values))
                
                if only_in_dict1:
                    print('Values only in source server:', list(only_in_dict1))
                
                if only_in_dict2:
                    print('Values only in destination server:', list(only_in_dict2))
                    
                print()

def populate_groups(f_bw_api_endpoint,f_access_token, f_group_list, f_coll_dict):
    
    coll_dict_id = {}
    for each_col in f_coll_dict:
        coll_dict_id[each_col["id"]] = each_col["name"]

    group_dict_name = {}
    for each_group in f_group_list:
        group_details = load_group_details_api(f_bw_api_endpoint,f_access_token, each_group["id"])
        group_dict_name[each_group["name"]] = []

        if "collections" in group_details.keys():
            for each_col_group in group_details["collections"]:
                if each_col_group["id"] in coll_dict_id.keys():
                    group_dict_name[each_group["name"]].append(coll_dict_id[each_col_group["id"]])

    return group_dict_name

def populate_members(f_bw_api_endpoint,f_access_token, f_member_list, f_coll_dict):
    
    coll_dict_id = {}
    for each_col in f_coll_dict:
        coll_dict_id[each_col["id"]] = each_col["name"]

    member_dict_email = {}
    for each_member in f_member_list:
        member_details = get_members_details(f_bw_api_endpoint, f_access_token, each_member["id"])
        member_dict_email[each_member["email"]] = []

        if "collections" in member_details.keys():
            for each_col_member in member_details["collections"]:
                if each_col_member["id"] in coll_dict_id.keys():
                    member_dict_email[each_member["email"]].append(coll_dict_id[each_col_member["id"]])

    return member_dict_email

def do_diff_collections(coll_dict_source, coll_dict_dest):

    #print(f"Number of collections: Source: {len(coll_dict_source)}, Destination: {len(coll_dict_dest)}")

    #extracting the names

    name_list_source = {}
    name_list_dest = {}
    
    for each_col in coll_dict_source:
        name_list_source[each_col["name"]]= []

    for each_col in coll_dict_dest:
        name_list_dest[each_col["name"]]= []
    
    compare_dicts(name_list_source, name_list_dest)

def do_diff_bw_to_bw():
    global access_token, dest_access_token, bw_cli_session, dest_bw_cli_session

    initial_environment_check()
    get_account_passwords()

    if access_token == "":
        access_token = login_to_bw_public_api(bw_identity_endpoint, bw_org_client_id, bw_org_client_secret)

    if dest_access_token == "":
        dest_access_token = login_to_bw_public_api(dest_bw_identity_endpoint, dest_bw_org_client_id, dest_bw_org_client_secret)

    #do_diff_users(access_token, dest_access_token)

    bw_cli_session = login_on_cli(bw_vault_uri, bw_acc_client_id, bw_acc_client_secret, bw_acc_password)
    coll_dict_source = load_collections_cli(bw_org_id, bw_cli_session)

    dest_bw_cli_session = login_on_cli(dest_bw_vault_uri, dest_bw_acc_client_id, dest_bw_acc_client_secret, dest_bw_acc_password)
    coll_dict_dest = load_collections_cli(dest_bw_org_id, dest_bw_cli_session)

    print()
    print("comparing collections")
    print("----------------------------")
    do_diff_collections(coll_dict_source, coll_dict_dest)
    print("----------------------------")
    print()

    #getting group list, populate, and compate
    group_list_source = load_groups_api(bw_api_endpoint, access_token)
    group_list_dest = load_groups_api(dest_bw_api_endpoint, dest_access_token)

    group_dict_by_name_source = populate_groups(bw_api_endpoint, access_token, group_list_source, coll_dict_source)
    group_dict_by_name_dest = populate_groups(dest_bw_api_endpoint, dest_access_token, group_list_dest, coll_dict_dest)

    print("comparing groups")
    print("----------------------------")
    compare_dicts(group_dict_by_name_source, group_dict_by_name_dest)
    print("----------------------------")
    print()

    #getting member list, populate, and compate
    member_list_source = get_members_list(bw_api_endpoint, access_token)
    member_list_dest = get_members_list(dest_bw_api_endpoint, dest_access_token)

    member_dict_by_name_source = populate_members(bw_api_endpoint, access_token, member_list_source, coll_dict_source)
    member_dict_by_name_dest = populate_members(dest_bw_api_endpoint, dest_access_token, member_list_dest, coll_dict_dest)

    print("comparing members")
    print("----------------------------")
    compare_dicts(member_dict_by_name_source, member_dict_by_name_dest)
    print("----------------------------")
    print()

def export_permissions():
    global bw_vault_uri, bw_acc_client_id, bw_acc_client_secret, bw_acc_password

    initial_environment_check()
    get_account_passwords()
    bw_cli_session = login_on_cli(bw_vault_uri, bw_acc_client_id, bw_acc_client_secret, bw_acc_password)

    coll_list, member_details_list = load_permissions_from_origin(bw_cli_session)
    save_json_to_file(coll_list, 'export_permissions.json')
    save_json_to_file(member_details_list, 'export_members.json')

def save_json_to_file(f_my_list, f_filename):
    json_data = json.dumps(f_my_list, indent=4)
    print("writing to file: " + script_location + f_filename)

    # write json data to a file
    try:
        with open(os.path.join( script_location, f_filename), 'w', encoding='utf-8') as json_file:
            json_file.write(json_data)
    except IOError as e:
        print(f'An error occurred while writing to the file: {e}')
    except Exception as e:
        print(f'An unexpected error occurred: {e}')

def print_help():
    print("usage: bwAdminTools.py <options>")
    print("")
    print("Options:")
    sys.stdout.write("%-18s %-50s\n" % ("-h, --help","Display help for commands "))
    sys.stdout.write("%-18s %-50s\n" % ("-c","The commands. See below for command list"))
    sys.stdout.write("%-18s %-50s\n" % ("-f, --config","File contains BW and LP configurations. Default: config.cfg"))
    print("")
    print("Commands:")
    sys.stdout.write("%-18s %-50s\n" % ("migratebw","To migrate from one Bitwarden server to another server"))
    sys.stdout.write("%-18s %-50s\n" % ("migratebwusers","To migrate users collection permission from one Bitwarden server to another server"))
    sys.stdout.write("%-18s %-50s\n" % ("diffbw","To compare collections, groups, and members between 2 organizations"))
    print("")
    print("Examples:")
    print("python3 bwAdminTools.py -c migratebw -f myconfig.cfg")
    print("python3 bwAdminTools.py -c migratebwusers -f myconfig.cfg ")

def load_configfile_lastpass(config):
    global bw_vault_mgmt_uri, lp_cid, lp_api_secret, lp_api_uri
    try:
        bw_vault_mgmt_uri = config.get('config', 'bw_vault_mgmt_uri')
        lp_cid = config.get('config', 'lp_cid')
        lp_api_secret = config.get('config', 'lp_api_secret')
        lp_api_uri = config.get('config', 'lp_api_uri')
    except configparser.NoSectionError as e:
        print(f"Missing 'config' section in the config file. Error: {str(e)}")
        sys.exit(2)
    except configparser.NoOptionError as e:
        print(f"Missing required option in the config file. Error: {str(e)}")
        sys.exit(2)
    except Exception as e:
        print(f"An unknown error occurred. Error: {str(e)}")
        sys.exit(2)
    if bw_vault_mgmt_uri[-1] != '/':
        bw_vault_mgmt_uri = bw_vault_mgmt_uri + "/"

    variables = {'bw_vault_mgmt_uri': bw_vault_mgmt_uri, 
                'lp_cid': lp_cid, 
                'lp_api_secret': lp_api_secret, 
                'lp_api_uri': lp_api_uri}

    empty_vars = [k for k, v in variables.items() if v == "" or v == '""']

    if empty_vars:
        print("Empty config found. Please check the config file.")
        for var in empty_vars:
            print(f"The variable '{var}' is empty.")
        sys.exit(2)


def load_configfile_bw2bw(config):
    global dest_bw_vault_uri, dest_bw_org_client_id, dest_bw_org_client_secret, dest_bw_acc_client_id, dest_bw_acc_client_secret, dest_bw_org_id, dest_bw_identity_endpoint, dest_bw_api_endpoint
    global bw_acc_client_id, bw_acc_client_secret

    try:
        bw_acc_client_id = config.get('config', 'bw_acc_client_id')
        bw_acc_client_secret = config.get('config', 'bw_acc_client_secret')
        dest_bw_vault_uri = config.get('config', 'dest_bw_vault_uri')
        dest_bw_org_client_id = config.get('config', 'dest_bw_org_client_id')
        dest_bw_org_client_secret = config.get('config', 'dest_bw_org_client_secret')
        dest_bw_acc_client_id = config.get('config', 'dest_bw_acc_client_id')
        dest_bw_acc_client_secret = config.get('config', 'dest_bw_acc_client_secret')
        dest_bw_org_id = config.get('config', 'dest_bw_org_id')

    except configparser.NoSectionError as e:
        print(f"Missing 'config' section in the config file. Error: {str(e)}")
        sys.exit(2)
    except configparser.NoOptionError as e:
        print(f"Missing required option in the config file. Error: {str(e)}")
        sys.exit(2)
    except Exception as e:
        print(f"An unknown error occurred. Error: {str(e)}")
        sys.exit(2)


    variables = {'dest_bw_vault_uri': dest_bw_vault_uri, 
                'dest_bw_org_client_id': dest_bw_org_client_id, 
                'dest_bw_org_client_secret': dest_bw_org_client_secret, 
                'dest_bw_acc_client_id': dest_bw_acc_client_id,
                'dest_bw_acc_client_secret' : dest_bw_acc_client_secret,
                'dest_bw_org_id' : dest_bw_org_id,
                'bw_acc_client_id' : bw_acc_client_id,
                'bw_acc_client_secret' : bw_acc_client_secret,
                }

    empty_vars = [k for k, v in variables.items() if v == "" or v == '""']

    if empty_vars:
        print("Empty config found. Please check the config file.")
        for var in empty_vars:
            print(f"The variable '{var}' is empty.")
        sys.exit(2)

    if dest_bw_vault_uri[-1] != '/':
        dest_bw_vault_uri = dest_bw_vault_uri + "/"

    if dest_bw_vault_uri == "https://bitwarden.com/":
        dest_bw_vault_uri == "https://vault.bitwarden.com/"
    elif dest_bw_vault_uri == "https://bitwarden.eu/":
        dest_bw_vault_uri == "https://vault.bitwarden.eu/"

    if dest_bw_vault_uri == "https://vault.bitwarden.com/":
        dest_bw_identity_endpoint = "https://identity.bitwarden.com/"
        dest_bw_api_endpoint = "https://api.bitwarden.com/"
    elif dest_bw_vault_uri == "https://vault.bitwarden.eu/":
        dest_bw_identity_endpoint = "https://identity.bitwarden.eu/"
        dest_bw_api_endpoint = "https://api.bitwarden.eu/"
    else:
        dest_bw_identity_endpoint = dest_bw_vault_uri + "identity/"
        dest_bw_api_endpoint = dest_bw_vault_uri + "api/"

def load_configfile_basic(config):
    global bw_vault_uri, bw_org_client_id, bw_org_client_secret, bw_identity_endpoint, bw_api_endpoint, bw_org_id
    try:
        bw_vault_uri = config.get('config', 'bw_vault_uri')
        bw_org_client_id = config.get('config', 'bw_org_client_id')
        bw_org_client_secret = config.get('config', 'bw_org_client_secret')
        bw_org_id = config.get('config', 'bw_org_id')

    except configparser.NoSectionError as e:
        print(f"Missing 'config' section in the config file. Error: {str(e)}")
        sys.exit(2)
    except configparser.NoOptionError as e:
        print(f"Missing required option in the config file. Error: {str(e)}")
        sys.exit(2)
    except Exception as e:
        print(f"An unknown error occurred. Error: {str(e)}")
        sys.exit(2)

    variables = {'bw_vault_uri': bw_vault_uri, 
                'bw_org_client_id': bw_org_client_id, 
                'bw_org_client_secret': bw_org_client_secret, 
                'bw_org_id': bw_org_id,
                }

    empty_vars = [k for k, v in variables.items() if v == "" or v == '""']

    if empty_vars:
        print("Empty config found. Please check the config file.")
        for var in empty_vars:
            print(f"The variable '{var}' is empty.")
        sys.exit(2)

    if bw_vault_uri[-1] != '/':
        bw_vault_uri = bw_vault_uri + "/"

    if bw_vault_uri == "https://bitwarden.com/":
        bw_vault_uri = "https://vault.bitwarden.com/"
    elif bw_vault_uri == "https://bitwarden.eu/":
        bw_vault_uri = "https://vault.bitwarden.eu/"

    if bw_vault_uri == "https://vault.bitwarden.com/":
        bw_identity_endpoint = "https://identity.bitwarden.com/"
        bw_api_endpoint = "https://api.bitwarden.com/"
    elif bw_vault_uri == "https://vault.bitwarden.eu/":
        bw_identity_endpoint = "https://identity.bitwarden.eu/"
        bw_api_endpoint = "https://api.bitwarden.eu/"
    else:
        bw_identity_endpoint = bw_vault_uri + "identity/"
        bw_api_endpoint = bw_vault_uri + "api/"

def load_configfile(configfile, command):
  
    config = configparser.ConfigParser()
    config.read(configfile)

    if 'config' not in config.sections():
        config.add_section('config')

    load_configfile_basic(config)
    if command in ["migratebw","migratebwusers","diffbw","exportperm","purgecoldest","purgegroupdest"]:
        load_configfile_bw2bw(config)
    elif command == "migratesf":
        load_configfile_lastpass(config)

def main(argv):
    configfile = "config.cfg"
    command = ""
    global access_token
    global verbose
    global debug

    try:
        opts, args = getopt.getopt(argv,"hvdc:f:",["help","config="])
    except getopt.GetoptError:
        print("Invalid options!")   
        print_help()
        sys.exit(2)
    for opt, arg in opts:
        if opt in ("-h", "--help"):
            print_help()
            sys.exit()
        elif opt in ("-f", "--config"):
            configfile = arg
        elif opt == "-c":
            command = arg
        elif opt == "-v":
            verbose = True
        elif opt == "-d":
            debug = True

    if os.path.exists(configfile):
        load_configfile(configfile, command)
    else:
        print("Config file is not found.")
        sys.exit(2)

    if command == "purgecol":
        delete_all_collections(bw_identity_endpoint, bw_api_endpoint, bw_org_client_id, bw_org_client_secret)
    elif command == "purgecoldest":
        delete_all_collections(dest_bw_identity_endpoint, dest_bw_api_endpoint, dest_bw_org_client_id, dest_bw_org_client_secret)
    elif command == "purgegroup":
        delete_all_groups(bw_identity_endpoint, bw_api_endpoint, bw_org_client_id, bw_org_client_secret)
    elif command == "purgegroupdest":
        delete_all_groups(dest_bw_identity_endpoint, dest_bw_api_endpoint, dest_bw_org_client_id, dest_bw_org_client_secret)
    elif (command == "migratesf") or (command == "migratelp"):
        migrate_shared_folders_permissions()
    elif command == "migratebw":
        migrate_data_bw_to_bw()
    elif command == "migratebwusers":
        migrate_users_bw_to_bw()
    elif command == "exportperm":
        export_permissions()
    elif command == "diffbw":
        do_diff_bw_to_bw()
        
    else:
        print("Invalid Command!")
        print_help()
        sys.exit(2)

if __name__ == "__main__":
    if len(sys.argv) > 1:
        main(sys.argv[1:])
    else:
        print_help()
